#lang racket
(require "test.rkt")

;; First try at a parser.

;; A table is a list of productions.
;; A production is a (cons sym list-of-case)
;;   where the sym is a non-terminal name
;; A case is a (list sequence thunk)
;;   where the thunk is the "action" for the case
;; A sequence is a list-of-item
;; An item is either a string (literal) or symbol (non-terminal)

(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]))

;; parse : sym table list-of-string -> any-or-false
(define (parse what table stream)
  (define defn (assq what table))
  (if defn
      (parse-alts (rest defn) table stream)
      (error 'parse "no definition for ~a" what)))

(test (parse 'main table '("apple")) 'apple)
(test (parse 'fruit-salad table '("orange" "orange")) 
      '(orange orange))
(test (parse 'main table '("radish")) #f)
#;(test (parse 'main table '("orange" "orange" "!")) 
        (cons '((orange orange) !) '()))

;; parse-alts : list-of-case table list-of-string -> any-or-false
(define (parse-alts alts table stream)
  (if (null? alts)
      false
      (or (parse-case (first alts) table stream)
          (parse-alts (rest alts) table stream))))

(let ()
  (define alts (list
                `[("apple") ,(lambda () 'apple)]
                `[("grape") ,(lambda () 'grape)]))
  (test (parse-alts alts table '("apple")) 'apple)
  (test (parse-alts alts table '("grape")) 'grape)
  (test (parse-alts alts table '("radish")) #f))

;; parse-case : case table list-of-string -> any-or-false
(define (parse-case case table stream)
  (define seq (first case))
  (define handler (second case))
  (define parsed (parse-sequence seq table stream))
  (and parsed
       (apply handler parsed)))

(let ()
  (define case `[("apple") ,(lambda () 'apple)])
  (test (parse-case case table '("apple")) 'apple)
  (test (parse-case case table '("grape")) #f))

;; parse-sequence : sequence table list-of-string -> list-or-false
(define (parse-sequence seq table stream)
  (cond
    [(empty? seq) empty]
    [else (local [(define first-parsed 
                    (parse-item (first seq) table stream))]
            (and first-parsed
                 (local [(define rest-parsed
                           (parse-sequence (rest seq) table stream))]
                   (and rest-parsed
                        (append first-parsed rest-parsed)))))]))

(let ()
  (define seq `("apple" "grape"))
  (test (parse-sequence '("apple") table '("apple")) empty)
  #;(test (parse-sequence seq table '("apple" "grape")) empty)
  #;(test (parse-sequence seq table '("apple" "radish")) #f))

;; parse-item : item table list-of-string -> list-or-false
(define (parse-item item table stream)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         empty
         false)]
    [else
     (define x (parse item table stream))
     (and x (list x))]))

(test (parse-item '"grape" table '("grape")) empty)
(test (parse-item 'fruit table '("grape")) '(grape))
(test (parse-item 'fruit table '("radish")) #f)

(run-tests)
